sync_schema_props_perms(syncprops=False)
if not ('require_permission' in schema and
        ('Repository', 'CWPermission') in schema['require_permission'].rdefs):
    add_relation_definition('Repository', 'require_permission', 'CWPermission')



#!/usr/bin/python
# -*- coding: utf-8 -*-

# A simple Mercurial hook to send update notifications to a
# cubicweb-vcsfile instance (via a broker, see notifbroker.py). As is,
# it makes the assumption that repos are organized under a 'root'
# directory called 'repos'. This is well suited for a
# Mercurial-server¹ based configuration.
#
# ¹: http://www.lshift.net/mercurial-server.html

import zmq
import os
import time

context = zmq.Context()
req_sock = context.socket(zmq.PUSH)
req_url = os.environ.get('HG_NOTIF_URL', 'tcp://127.0.0.1:5556')
req_sock.connect(req_url)

cwd = os.environ['PWD']
msg = cwd.split('/repos/')[-1].rstrip('/')
req_sock.send(msg)

# make sure the socket has some time to send the payload
time.sleep(0.5)
# make sure we won't block on exit event is there is no listening server
req_sock.close(linger=0)

print 'Sent ZMQ notification for', msg, 'to', req_url



"""custom forms to submit new revision to the svn repository or to edit
some information about existing revisions.

:organization: Logilab
:copyright: 2007-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from copy import copy

from logilab.mtconverter import xml_escape
from logilab.common.decorators import cached

from cubicweb.predicates import (match_form_params, match_kwargs,
                                 is_instance, specified_etype_implements,
                                 match_edited_type)
from cubicweb import Forbidden
from cubicweb.view import EntityView
from cubicweb.web import (Redirect, RequestError, controller, uicfg, form,
                          formfields as ff, formwidgets as fw)
from cubicweb.web.views import autoform, forms, editcontroller

from cubes.vcsfile import IMMUTABLE_ATTRIBUTES

_afs = uicfg.autoform_section
_aff = uicfg.autoform_field
_affk = uicfg.autoform_field_kwargs
_rctl = uicfg.reledit_ctrl
_pvdc = uicfg.primaryview_display_ctrl

for attr in ('path', 'subpath', 'type', 'source_url', 'encoding'):
    _pvdc.tag_attribute(('Repository', attr), {'vid': 'attribute'})
_rctl.tag_attribute(('Repository', 'title'), {'reload': True})
_affk.tag_attribute(('Repository', 'source_url'), {'widget': fw.TextInput})
_affk.tag_attribute(('Repository', 'path'), {'widget': fw.TextInput})
_affk.tag_attribute(('Repository', 'subpath'), {'widget': fw.TextInput})
_afs.tag_attribute(('Repository', 'local_cache'), 'main', 'hidden')

_afs.tag_attribute(('Revision', 'revision'), 'main', 'hidden')
_afs.tag_attribute(('Revision', 'changeset'), 'main', 'hidden')
_affk.tag_attribute(('Revision', 'author'),
                    {'value': lambda form, field: form._cw.user.login})
_afs.tag_subject_of(('Revision', 'parent_revision', 'Revision'), 'main', 'hidden')
_afs.tag_object_of(('Revision', 'parent_revision', 'Revision'), 'main', 'hidden')

_affk.tag_attribute(('VersionedFile', 'directory'), {'widget': fw.TextInput})
_affk.tag_attribute(('VersionedFile', 'name'), {'widget': fw.TextInput})
_afs.tag_object_of(('*', 'content_for', 'VersionedFile'), 'main', 'hidden')

_afs.tag_object_of(('*', 'from_revision', 'Revision'), 'main', 'inlined')

_afs.tag_subject_of(('*', 'content_for', 'VersionedFile'), 'main', 'hidden')
_aff.tag_attribute(('VersionContent', 'data'), ff.EditableFileField)


class RevisionCreationForm(autoform.AutomaticEntityForm):
    # vfeid needed for the inline-creation of VersionContent
    __select__ = specified_etype_implements('Revision') & match_form_params('vfeid')

    def should_display_inline_creation_form(self, rschema, existant, card):
        if rschema == 'from_revision':
            return True
        return super(RevisionCreationForm, self).should_display_inline_creation_form(
            rschema, existant, card)


class RevisionEditController(editcontroller.EditController):
    __select__ = editcontroller.EditController.__select__ & match_edited_type('Revision')

    def _insert_entity(self, etype, eid, rqlquery):
        try:
            repoeid = self.linked_to[('from_repository', 'subject')][0]
        except KeyError:
            raise RequestError('missing repository information')
        self._cw.set_shared_data('vcsrepoeid', repoeid, querydata=True)
        return super(RevisionEditController, self)._insert_entity(etype, eid, rqlquery)

class VCInlinedCreationFormView(autoform.InlineEntityCreationFormView):
    # vfeid needed for the inline-creation of VersionContent
    __select__ = (match_form_params('vfeid') & match_kwargs('peid', 'rtype')
                  & specified_etype_implements('VersionContent'))

    @cached
    def _entity(self):
        """creation view for an entity"""
        vf = self._cw.entity_from_eid(int(self._cw.form['vfeid']))
        self.orig_entity = vf.head
        # need a fully completed entity before doing the copy
        self.orig_entity.complete()
        if self.orig_entity.attr_metadata('data', 'format') in (
            'text/plain', 'text/html', 'text/rest'):
            # fill cache, Bytes fields are not loaded by complete()
            self.orig_entity.data
        entity = copy(self.orig_entity)
        self.initialize_varmaker()
        entity.eid = self.varmaker.next()
        return entity

    def add_hiddens(self, form, entity):
        super(VCInlinedCreationFormView, self).add_hiddens(
            form, entity)
        form.add_hidden('content_for', self.orig_entity.file.eid,
                        eidparam=True)


def filter_out_immutable_rels(func):
    def wrapper(self, *args, **kwargs):
        entity = self.edited_entity
        if entity.has_eid():
            immutable = IMMUTABLE_ATTRIBUTES
        else:
            immutable = ('Revision.revision',)
        for rschema, tschemas, role in func(self, *args, **kwargs):
            if role == 'subject':
                etype = entity.__regid__
            # take care, according to the wrapped function tschemas is a list or
            # a single schema
            elif isinstance(tschemas, list):
                etype = tschemas[0]
            else:
                etype = tschemas
            if '%s.%s' % (etype, rschema) in immutable:
                continue
            yield rschema, tschemas, role
    return wrapper


class VCRevisionEditionForm(autoform.AutomaticEntityForm):
    __select__ = is_instance('Revision', 'VersionedFile',
                            'VersionContent', 'DeletedVersionContent')
    _relations_by_section = filter_out_immutable_rels(
        autoform.AutomaticEntityForm._relations_by_section)


def available_branches(form, field):
    return form.edited_entity.repository.branches()


class NewRevisionForm(forms.EntityFieldsForm):
    __regid__ = 'vfnewrevform'
    __select__ = is_instance('VersionedFile')

    form_renderer_id = 'base'
    form_buttons = [fw.SubmitButton()]

    file = ff.FileField(eidparam=True)
    branch = ff.StringField(required=True, eidparam=True,
                            widget=fw.Select, choices=available_branches,
                            value=lambda form:form.edited_entity.repository.default_branch())
    msg = ff.StringField(label=_('commit message'), eidparam=True)


class VFUploadFormView(form.FormViewMixIn, EntityView):
    """form to upload a new revision of a versioned file"""
    __regid__ = 'vfnewrevform'
    __select__ = is_instance('VersionedFile')

    submitmsg = _('new revision has been checked-in')

    @property
    def action(self):
        return self._cw.build_url('vcsnewrev')

    def call(self, **_kwargs):
        # XXX explain why we would want self.rset.rowcount != 1
        for i in xrange(self.cw_rset.rowcount):
            self.cell_call(i, 0)

    def form_title(self, entity):
        return self._cw._('Upload a new revision of %s') % (
            '<a href="%s">%s</a>' % (xml_escape(entity.absolute_url()),
                                     xml_escape(entity.dc_title())))

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<div class="iformTitle">')
        self.w(self.form_title(entity))
        self.w(u'</div>')
        self.get_form(entity).render(w=self.w)

    def get_form(self, entity):
        return self._cw.vreg['forms'].select(
            'vfnewrevform', self._cw, entity=entity,
            submitmsg=self._cw._(self.submitmsg), action=self.action,
            __redirectpath=entity.rest_path(),)


class VFUploadController(controller.Controller):
    """upload a new revision of an already existing document"""
    __regid__ = 'vcsnewrev'

    def publish(self, rset=None):
        for eid in self._cw.edited_eids():
            formparams = self._cw.extract_entity_params(eid, minparams=2)
            vf = self._cw.entity_from_eid(eid)
            vf.complete()
            self.upload_revision(vf, formparams)
        try:
            kwargs = {'__message': self._cw.form['__message']}
        except KeyError:
            kwargs = {}
        try:
            goto = self._cw.build_url(self._cw.form['__redirectpath'], **kwargs)
        except KeyError:
            goto = vf.absolute_url(**kwargs)
        raise Redirect(goto)

    def upload_revision(self, vf, formparams):
        try:
            # [-1] to discard file name and mime type
            stream = formparams['file'][-1]
            vf.vcs_upload_revision(stream, branch=formparams['branch'],
                                   msg=formparams['msg'])
            return vf.eid
        except KeyError:
            raise Forbidden('Invalid content')


class VcsRmForm(form.FormViewMixIn, EntityView):
    __regid__ = 'vcsrmform'
    __select__ = is_instance('VersionedFile')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<div class="iformTitle">')
        self.w(self._cw._('DELETE the versioned file %(file)s') %
                    {'file': entity.view('incontext')})
        self.w(u'</div>')
        # XXX branch
        form = self._cw.vreg['forms'].select('base', self._cw, rset=self.cw_rset,
                                             form_renderer_id='base', entity=entity,
                                             action=self._cw.build_url('vcsrm'),
                                             form_buttons=[fw.SubmitButton()])
        form.append_field(ff.StringField(name='msg', eidparam=True,
                                         value=u'',
                                         label=_('commit message')))
        form.render(w=self.w)


class VcsRmController(controller.Controller):
    """DELETE"""
    __regid__ = 'vcsrm'

    def publish(self, rset=None):
        req = self._cw
        for eid in req.edited_eids():
            formparams = req.extract_entity_params(eid, minparams=2)
            vf = self._cw.entity_from_eid(eid)
            vf.vcs_rm(msg=formparams['msg'])
        msg = req._('file was marked DELETED')
        raise Redirect(vf.absolute_url(__message=msg))



# -*- coding: utf-8 -*-
"""makes the bridge between repository handlers and related cubicweb entities

:organization: Logilab
:copyright: 2007-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from __future__ import with_statement

__docformat__ = "restructuredtext en"

import sys
from os.path import join

from logilab.common.modutils import LazyObject
from logilab.mtconverter import guess_mimetype_and_encoding
from yams import ValidationError

from cubicweb import Binary, QueryError, typed_eid
from cubicweb.server.session import security_enabled
from cubes.vcsfile import queries

_VCSTYPES = {'subversion': LazyObject('cubes.vcsfile.reposvn', 'SVNRepository'),
             'mercurial': LazyObject('cubes.vcsfile.repohg', 'HGRepository'),
             }
_REPOHDLRS = {}


class VCSException(Exception):
    def __init__(self, repoeid, attr, msgid, msgargs=()):
        self.repoeid = repoeid
        self.attr = attr
        self.msgid = msgid
        self.msgargs = msgargs

    def __str__(self):
        return '%s.%s: %s' % (self.repoeid, self.attr,
                              self.msgid % self.msgargs)

    def to_validation_error(self, translate):
        msg = translate(self.msgid)
        if self.msgargs:
            msg %= self.msgargs
        return ValidationError(self.repoeid, {self.attr: msg})


def repository_handler(repoent):
    """get repository handler for the given repository entity"""
    # deactivate read security first, potentially needed to get access to
    # path/local_cache
    session = repoent._cw
    with security_enabled(session, read=False):
        try:
            repohdlr = _REPOHDLRS[repoent.eid]
            repohdlr.encoding = repoent.encoding # in case it changed
            return repohdlr
        except KeyError:
            try:
                path = repoent.path
                if not path:
                    if not repoent.local_cache:
                        msg = session._('no local access to repository')
                        raise ValidationError(repoent.eid, {'source_url': msg})
                    path = join(session.vreg.config.local_repo_cache_root,
                                repoent.local_cache)
                repohdlr = _VCSTYPES[repoent.type](repoent.eid, path,
                                                   repoent.encoding)
            except KeyError:
                msg = session._('%s is not a known repository type')
                raise ValidationError(repoent.eid, {'type': msg % repoent.type})
            except ImportError:
                msg = session._('missing python bindings to support %s repositories')
                raise ValidationError(repoent.eid, {'type': msg % repoent.type})
            _REPOHDLRS[repoent.eid] = repohdlr
            return repohdlr


def import_content(session, commitevery=1000, raise_on_error=False):
    """synchronize content of known vcs repositories
    """
    for vcsrepo in session.execute(
        'Any X,T,U,P,L,SP,E WHERE X is Repository, '
        'X type T, X source_url U, X path P, X local_cache L, '
        'X subpath SP, X encoding E').entities():
        try:
            repohdlr = repository_handler(vcsrepo)
        except VCSException, ex:
            if raise_on_error:
                raise
            session.error(str(ex))
            continue
        import_vcsrepo_content(session, repohdlr, vcsrepo,
                               commitevery=commitevery,
                               raise_on_error=raise_on_error)


def import_vcsrepo_content(session, repohdlr, vcsrepo, commitevery=10,
                           raise_on_error=False):
    """synchronize content of given vcs repository"""
    try:
        repohdlr.import_content(vcsrepo, commitevery)
        # give other threads a chance to run: commit will release the cnxset, get
        # it back afterwards
        session.commit()
    except Exception:
        if raise_on_error:
            raise
        # clear in cases new files have been added in cache, but entities will
        # be rollbacked
        repohdlr.vf_path_cache.clear()
        try:
            vcsrepo._cw.set_cnxset() # ensure a cnxset is set
            title = vcsrepo.dc_title()
        except Exception:
            title = str(vcsrepo)
        repohdlr.exception(
            'error while importing content for vcs repo %s', title)
        session.rollback()
    session.set_cnxset()


def set_at_revision(session, reid, parent_lrevs=None, safetybelt=False):
    """at_revision is an internal relation used to know the full content of a
    repository at any revision. This function set this relation for newly
    created revision.

    `parent_lrevs` should be an ordered list of parent revision, the first
    one being the one into which others are merged
    """
    sql = session.system_sql
    if safetybelt:
        sql("DELETE FROM at_revision_relation WHERE eid_to='%s'" % reid)
    # link to version content generated by this revision
    # rql = 'SET VC at_revision REV WHERE REV eid %(rev)s, VC from_revision REV'
    sqlrestr = ('SELECT VC.cw_eid, %s FROM cw_%%s AS VC '
                'WHERE VC.cw_from_revision=%s' % (reid, reid))
    sql('INSERT INTO at_revision_relation(eid_from,eid_to) '
        '%s UNION %s' % (sqlrestr % 'VersionContent',
                         sqlrestr % 'DeletedVersionContent'))
    # link to version content of files in parent revisions but not modified by
    # this revision
    # rql = ('SET VC at_revision REV WHERE VC is VersionContent, '
    #        'REV eid %(rev)s, VC at_revision PREV, '
    #        'REV parent_revision PREV, '
    #        'NOT EXISTS(VC2 from_revision REV, VC content_for VF, '
    #        'VC2 content_for VF)')
    if parent_lrevs is None or len(parent_lrevs) == 1:
        sql('INSERT INTO at_revision_relation(eid_from,eid_to) '
            'SELECT DISTINCT VC.cw_eid, %(reid)s '
            'FROM at_revision_relation AS ar, parent_revision_relation AS pr, cw_VersionContent AS VC '
            'WHERE VC.cw_eid=ar.eid_from AND pr.eid_from=%(reid)s AND pr.eid_to=ar.eid_to '
            'AND NOT EXISTS('
            '  SELECT 1 FROM cw_VersionContent AS VC2 '
            '  WHERE VC2.cw_from_revision=%(reid)s AND VC2.cw_content_for=VC.cw_content_for '
            '  UNION '
            '  SELECT 1 FROM cw_DeletedVersionContent AS VC2 '
            '  WHERE VC2.cw_from_revision=%(reid)s AND VC2.cw_content_for=VC.cw_content_for)'
            % {'reid': reid})
    elif len(parent_lrevs) > 1:
        # getting back files "at_revision" parent revision from the first parent
        # revision is enough: other changes get back from merged parents as
        # version content "from_revision" the current revision
        sql('INSERT INTO at_revision_relation(eid_from,eid_to) '
            'SELECT DISTINCT VC.cw_eid, %(reid)s '
            'FROM at_revision_relation AS ar, parent_revision_relation AS pr, '
            '     cw_VersionContent AS VC, cw_Revision as R '
            'WHERE VC.cw_eid=ar.eid_from AND pr.eid_from=%(reid)s AND pr.eid_to=ar.eid_to '
            ' AND pr.eid_to=R.cw_eid AND R.cw_revision=%(plrev)s '
            ' AND NOT EXISTS('
            '  SELECT 1 FROM cw_VersionContent AS VC2 '
            '  WHERE VC2.cw_from_revision=%(reid)s AND VC2.cw_content_for=VC.cw_content_for '
            '  UNION '
            '  SELECT 1 FROM cw_DeletedVersionContent AS VC2 '
            '  WHERE VC2.cw_from_revision=%(reid)s AND VC2.cw_content_for=VC.cw_content_for)'
            % {'reid': reid, 'plrev': parent_lrevs.pop(0)})


def import_revision(session, repoeid, date, **kwargs):
    """import a new revision from a repository"""
    args = {'date': date}
    args['repoeid'] = repoeid
    for attr in ('revision', 'author', 'description', 'changeset', 'branch', 'phase', 'hidden'):
        args[attr] = kwargs.get(attr)
    prevs = kwargs['parents']
    if prevs:
        args['parent'] = prevs[0]
    precs = kwargs['precursors']
    if precs:
        args['precursor'] = precs[0]
    reveid = session.execute(queries.new_revision_rql(prevs, precs, True), args)[0][0]
    # complete other parents & precursors relations
    if len(prevs) > 1:
        for preveid in prevs[1:]:
            session.execute('SET R parent_revision PR WHERE R eid %(r)s, PR eid %(pr)s',
                            {'r': reveid, 'pr': preveid})
    if len(precs) > 1:
        for preceid in precs[1:]:
            session.execute('SET R obsoletes PREC WHERE R eid %(r)s, PREC eid %(pr)s',
                            {'r': reveid, 'pr': preceid})
    return reveid

def import_versioned_file(session, repoeid, date, directory, name):
    """import a version controlled file from a repository"""
    args = {'directory': directory, 'name': name, 'repoeid': repoeid}
    return session.execute(
        'INSERT VersionedFile X: X directory %(directory)s, '
        'X name %(name)s, X from_repository R '
        'WHERE R eid %(repoeid)s', args)[0][0]

def import_version_content(session, repoeid, reveid, upath, date, **kwargs):
    """import a new file revision from a repository"""
    repohdlr = repository_handler(session.entity_from_eid(repoeid, 'Repository'))
    vfeid = repohdlr.vf_eid(session, upath, date)
    # done in a hook but doing it here as well to avoid an additional query to
    # lookup filename
    mt, enc = guess_mimetype_and_encoding(
        fallbackencoding=repohdlr.encoding, fallbackmimetype=None,
        filename=upath)
    args = {'vf': vfeid, 'rev': reveid, 'data': kwargs['data'],
            'df': mt and unicode(mt) or None,
            'de': enc and unicode(enc) or None}
    rql = ('INSERT VersionContent X: X from_revision REV, X content_for VF, '
           'X data_format %(df)s, X data_encoding %(de)s, X data %(data)s')
    for inlinerel in ('vc_copy', 'vc_rename'):
        if inlinerel in kwargs:
            args[inlinerel] = kwargs[inlinerel]
            rql += ', X %s %%(%s)s' % (inlinerel, inlinerel)
    rql += ' WHERE REV eid %(rev)s, VF eid %(vf)s'
    return session.execute(rql, args)[0][0], vfeid


def import_deleted_version_content(session, repoeid, reveid, upath, date):
    """import a new deleted file revision from a repository"""
    repohdlr = repository_handler(session.entity_from_eid(repoeid, 'Repository'))
    vfeid = repohdlr.vf_eid(session, upath, date)
    args = {'vf': vfeid, 'rev': reveid}
    rql = ('INSERT DeletedVersionContent X: X from_revision REV, X content_for VF '
           ' WHERE REV eid %(rev)s, VF eid %(vf)s')
    return session.execute(rql, args)[0][0], vfeid


from cubicweb import QueryError
from cubicweb.server.sources import storages

class VCSStorage(storages.Storage):
    """store Bytes attribute value in a version content system"""

    def callback(self, source, session, value):
        repoeid, dir, name, revnum = source.binary_to_str(value).split('\x01')
        repoeid = typed_eid(repoeid)
        try:
            vcsrepohdlr = repository_handler(session.entity_from_eid(repoeid, 'Repository'))
        except Exception, ex:
            source.critical("can't get repository handler for eid %s: %s",
                            repoeid, ex, exc_info=sys.exc_info())
            return None
        try:
            return vcsrepohdlr.file_content(dir, name, revnum)
        except Exception, ex:
            source.critical("can't get content for %s/%s at revision %s in vcs "
                            "repository %s: %s",
                            dir, name, revnum, vcsrepohdlr, ex,
                            exc_info=sys.exc_info())
            return None

    def entity_added(self, entity, attr):
        """an entity using this storage for attr has been added"""
        try:
            data = entity.cw_edited.pop(attr)
        except KeyError:
            pass
        else:
            session = entity._cw
            # don't create vcs transaction if the revision is being imported
            # from the vcs repository
            try:
                repoeid, dir, name, revnum = session.transaction_data['vcs_importing']
            except KeyError:
                vf = entity._vc_vf()
                repoeid = vf.repository.eid
                vcsrepohdlr, transaction = entity._vc_prepare(repoeid)
                dir = vcsrepohdlr.encode(vf.directory)
                name = vcsrepohdlr.encode(vf.name)
                revnum = transaction.rev
                vcsrepohdlr.add_versioned_file_content(session, transaction,
                                                       vf, entity, data)
            vckey = (str(repoeid), dir, name, str(revnum))
            entity.cw_edited[attr] = Binary('\x01'.join(vckey))
            return data

    def entity_updated(self, entity, attr):
        """an entity using this storage for attr has been updatded"""
        # XXX create a new revision to make this storage usable for arbitrary
        # attributes
        raise QueryError('read-only attribute')

    def entity_deleted(self, entity, attr):
        """an entity using this storage for attr has been deleted"""
        # XXX delete from repository to make this storage usable for arbitrary
        # attributes (unless vcs-importing, meaning a strip has been detected)
        pass
